
/*
 * Copyright 2007 Design of Systems on Silicon (DS2) All rights reserved.
 */

/**
 * @addtogroup des
 * @{
**/

/**
 * @file des_pub.c
 * @brief This code implements the DES data-encryption algorithm.
 *
 * @internal
 *
 * @author Ramon Rodriguez
 * @date 29/03/2007
 * 
**/

#include <string.h>

#include "types.h"
#ifndef TOOLS
#  include "LCAssert_pub.h"
#else
#  include <assert.h>
#  define assertpoint(E, ID) (assert((E)))
#endif
#include "des_pub.h"
#include "des_protocol_pub.h"

void DES3CBCEncipher(INT8U *data, INT32U length, const INT8U *key)
{
  // INT8U key[32]; // 256 bits
  INT8U           iv[8];
  des3_context    ctx;

  assertpoint(length <= (INT32U)MAX_INT32S, 1);
  assertpoint((length % 8) == 0, 2);

  des3_set_3keys(&ctx, (unsigned char *)key);
  memcpy(iv, key + 24, sizeof(iv));
  des3_cbc_encrypt(&ctx, (unsigned char *)iv, (unsigned char *)data, (unsigned char *)data,
                   (int)length);
}

void DES3CBCDecipher(INT8U *data, INT32U length, const INT8U *key)
{
  // INT8U key[32]; // 256 bits
  INT8U           iv[8];
  des3_context    ctx;

  assertpoint(length <= (INT32U)MAX_INT32S, 3);
  assertpoint((length % 8) == 0, 4);

  des3_set_3keys(&ctx, (unsigned char *)key);
  memcpy(iv, key + 24, sizeof(iv));
  des3_cbc_decrypt(&ctx, (unsigned char *)iv, (unsigned char *)data, (unsigned char *)data,
                   (int)length);
}

INT8U DESComponentIsPresent(void)
{
  return 1;
}

/**
 * @}
**/
