
/*
 * Copyright 2007 Design of Systems on Silicon (DS2) All rights reserved.
 */

/**
 * @addtogroup des
 * @{
**/

/**
 * @file des_pub.h
 * @brief DES data-encryption algorithm header file.
 *
 * @internal
 *
 * @author Ramon Rodriguez
 * @date 29/03/2007
 *  
**/

#ifndef DES_PUB_H
#  define DES_PUB_H

/**
 * @brief Triple-DES-CBC buffer encryption (EDE3)
 *
 * @param data   buffer holding the data (plaintext input/ciphertext output)
 * @param length length of the input data
 * @param key    32-byte secret key, made from the concatenation of:
 *                 - k1: DES key 1 (encryption) (8 bytes)
 *                 - k2: DES key 2 (decryption) (8 bytes)
 *                 - k3: DES key 3 (encryption) (8 bytes)
 *                 - iv: initialization vector (8 bytes)
 *
 * @remarks data buffer must be an integer number of 8-byte data blocks
 *          length must be multiple of 8
 **/
void            DES3CBCEncipher(INT8U *data, INT32U length, const INT8U *key);

/**
 * @brief Triple-DES-CBC buffer decryption (EDE3)
 *
 * @param data   buffer holding the data (ciphertext input/plaintext output)
 * @param length length of the input data
 * @param key    32-byte secret key, made from the concatenation of:
 *                 - k1: DES key 1 (decryption) (8 bytes)
 *                 - k2: DES key 2 (encryption) (8 bytes)
 *                 - k3: DES key 3 (decryption) (8 bytes)
 *                 - iv: initialization vector (8 bytes)
 *
 * @remarks data buffer must be an integer number of 8-byte data blocks
 *          length must be multiple of 8
 **/
void            DES3CBCDecipher(INT8U *data, INT32U length, const INT8U *key);

/**
 * @brief Find out whether the component is Present or not
 *
 * @return 1 -> if is present 0->if not  
 **/
INT8U           DESComponentIsPresent(void);

#endif /* DES_PUB_H */

/**
 * @}
**/
