
/*
 * Copyright (c) 1985, 1989 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *  This product includes software developed by the University of
 *  California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *  from: @(#)ftp_var.h 5.9 (Berkeley) 6/1/90
 */

/**
 * @addtogroup ftp
 * @{
 * */

/**
 * @file ftp_var.h
 * @brief Declarations of File Transfer Protocol
 *
 * @author Copyright (c) 1985, 1989 Regents of the University of California.All rights reserved.
 *
 * @note From: @(#)ftp.c  5.38 (Berkeley) 4/22/91
 *
 * */

/// FTP global variables.

#ifndef __FTP_VAR__
#  define __FTP_VAR__

#  include <sys/param.h>
#  include "ftp_constants.h"

/// Tick counter step size.
//#define TICKBYTES     10240

#  define CTRL_TIMEOUT  10
#  define DATA_TIMEOUT  30
#  define MAX_RETRIES 2
#  define MAX_WRITE_RETRIES MAX_RETRIES
#  define MAX_READ_RETRIES  MAX_RETRIES

#  define MAX_BUF  256

/// Global variables. 

//int hash;   /* print # for each buffer transferred */
//extern int  connected;  /* connected to server */
//int ntflag;   /* use ntin ntout tables for name xlation */
//int mapflag;  /* use mapin mapout templates on file names */
//extern int  code;   /* return/reply code for ftp command */
//extern int  crflag;   /* if 1, strip car. rets. on ascii gets */
//char     pasv[64];       /* passive port for proxy data connection */
//char  *altarg;  /* argv[1] with no shell-like preprocessing  */
//char  ntin[17]; /* input translation table */
//char  ntout[17];  /* output translation table */
//char  mapin[MAXPATHLEN];  /* input map template */
//char  mapout[MAXPATHLEN]; /* output map template */
//char  typename[32];   /* name of file transfer type */
//extern int  type;     /* requested file transfer type */
//extern int  curtype;    /* current file transfer type */
//char  structname[32];   /* name of file transfer structure */
//int mode;     /* file transfer mode */
//char  bytename[32];   /* local byte size in ascii */

//extern char *hostname;  /* name of host connected to */
//int unix_server;  /* server is unix, can use binary for ascii */

//int ftp_port; /* htons'd port number for ftp service */

//extern int  cpend;    /* flag: if != 0, then pending server reply */
//int mflag;    /* flag: if != 0, then active multi command */

//int options;  /* used during socket creation */
//char reply_string[BUFSIZ];    /* last line of previous reply */
//char lreply[BUFSIZ];

struct types
{
  const char     *t_name;
  const char     *t_mode;
  int             t_type;
  const char     *t_arg;
};

/**
 * @brief Opens a socket connection to a host with a specific port. 
 *
 * @param host The host name
 * @param port The port
 *
 * @return The host name
 *
 **/
char           *hookup(char *host, int port);

/**
 * @brief Login the connection. 
 *
 * @param luser User name
 * @param passwd Password
 *
 * @return 0 Success, 1 Failed
 *
 **/
int             dologin(const char *luser, const char *passwd);

/**
 * @brief Executes the ftp commands
 * 
 * @param cmd The command
 *
 * @return int
 *
 **/
int             command(const char *cmd);

/**
 * @brief Sends a command request
 *
 * @param cmd The command
 * @param remote Remote Server
 * @param fct callback function
 *
 * @return int
 *
 **/
int             sendrequest(const char *cmd, char *remote, int (*fct) (char *, unsigned), unsigned,
                            off_t);

/**
 * @brief Process the response.
 *
 * @param cmd The command
 * @param remote Remote Server
 * @param fct Callback function
 *
 * @return int
 *
 **/
int             recvrequest(const char *cmd, char *remote, int (*fct) (char *, unsigned), unsigned,
                            off_t);

/**
 * @brief get an FTP command reply
 *
 * @param expecteof TRUE = EOF expected, FALSE = EOF is error
 *
 * @return If EOF encountered but not expected return 0
 *
 **/
int             getreply(int expecteof);

/**
 * @brief Changes type
 *
 * @param newtype ASCII ,EBCDIC, image, or local byte size
 * @param show Verbose mode
 *
 * @return int
 *
 **/
void            changetype(int newtype, int show);

/**
 * @brief Called by TCP Server to return a completed connection
 *
 * @return 0 : Success. New socket created ,-1: Error
 *
 **/
int             dataconn();

/**
 * @brief Start a listen on the data channel 
 *
 * @return 0 Fails, 1 Success
 *
 **/
int             initconn(void);

/**
 * @brief Verbose information about transfer (bytes/sec)
 *
 * @param direction string direction
 * @param bytes Number of bytes
 * @param t0 Time operator 0
 * @param t1 Time operator 1
 *
 * @return void
 *
 **/
void            ptransfer(const char *direction, long bytes, const struct timeval *t0,
                          const struct timeval *t1);

/**
 * @brief Substract time structures (t1 - t0) 
 *
 * @param tdiff Time structure resulted
 * @param t1 Operator 1
 * @param t0 Operator 0
 *
 * @return void
 *
 **/
void            tvsub(struct timeval *tdiff, const struct timeval *t1, const struct timeval *t0);

/**
 * @brief  Shuts down part of a connection
 *
 **/
void            lostpeer(int);

/**
 * @brief Obtains the log level 
 *
 * @return Log level
 *
 **/
t_log          *ftpLogGet(void);

#endif

/**
 * @}
 * */
