
/**
 * @addtogroup ftp
 * @{
 * */

/**
 * @file ftpapi.c
 * @brief API of the FTP client
 *
 * @author Francisco Escuder
 *
 * @note 
 *
 * */

#include "ftp_pub.h"
#include "ftplib.h"
#include "ftp_constants.h"

//static pthread_mutex_t G_lowLevel_mutex;
//static pthread_mutex_t G_highLevel_mutex;

t_ftpError FtpOpen(char *host, char *login, char *passwd)
{
  if ((host == NULL) || (login == NULL) || (passwd == NULL))
    return -1;
  //  pthread_mutex_init(&G_lowLevel_mutex, NULL);
  //  pthread_mutex_init(&G_highLevel_mutex, NULL);
  return ftp_open(host, login, passwd);
}

t_ftpError FtpClose(void)
{
  //  pthread_mutex_destroy(&G_lowLevel_mutex);
  //  pthread_mutex_destroy(&G_highLevel_mutex);
  return ftp_close();
}

t_ftpError FtpGetFile(char *remote_name, char *local_name)
{
  int             rc = 0;

  //  pthread_mutex_lock (&G_highLevel_mutex);
  if ((remote_name == NULL) || (local_name == NULL))
    return FTP_ERR_INVALID_ARGUMENT;
  else
    rc = ftp_getfile(remote_name, local_name, 0);
  //  pthread_mutex_unlock (&G_highLevel_mutex);
  if (rc)
  {
    if (rc == FTP_FILE_ERROR)
      return FTP_ERR_FILE_ERROR;
    return FTP_ERR_DOWNLOAD_ERROR;
  }
  return rc;
}

t_ftpError FtpPutFile(char *local_name, char *remote_name, int append)
{
  int             rc = 0;

  //  pthread_mutex_lock (&G_highLevel_mutex);
  if ((remote_name == NULL) || (local_name == NULL))
    return FTP_ERR_INVALID_ARGUMENT;
  else
    rc = ftp_putfile(local_name, remote_name, 0, append);
  //  pthread_mutex_unlock (&G_highLevel_mutex);
  if (rc)
  {
    if (rc == FTP_FILE_ERROR)
      return FTP_ERR_FILE_ERROR;
    return FTP_ERR_DOWNLOAD_ERROR;
  }
  return rc;
}

t_ftpError FtpGet(char *remote_name, int (*cons) (char *, unsigned), unsigned bufsize)
{
  int             rc = 0;

  //  pthread_mutex_lock (&G_lowLevel_mutex);
  rc = ftp_get(remote_name, cons, bufsize, 0);
  //pthread_mutex_unlock (&G_lowLevel_mutex);

  if (rc)
  {
    if (rc == FTP_FILE_ERROR)
      return FTP_ERR_FILE_ERROR;
    return FTP_ERR_DOWNLOAD_ERROR;
  }
  return rc;
}

t_ftpError FtpPut(char *remote_name, int (*prod) (char *, unsigned), unsigned bufsize, int append)
{
  int             rc = 0;

  //pthread_mutex_lock (&G_lowLevel_mutex);
  rc = ftp_put(remote_name, prod, bufsize, 0, append);
  //pthread_mutex_unlock (&G_lowLevel_mutex);
  if (rc)
  {
    if (rc == FTP_FILE_ERROR)
      return FTP_ERR_FILE_ERROR;
    return FTP_ERR_DOWNLOAD_ERROR;
  }
  return rc;
}

t_ftpError FtpLs(int long_ls, char *file, int (*cons) (char *, unsigned))
{
  return ftp_ls(long_ls, file, cons);
}

/**
 * @}
 * */
