
/* Copyright (c) 1985, 1989 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *  This product includes software developed by the University of
 *  California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/**
 * @addtogroup ftp
 * @{
 * */

/**
 * @file ftplib.c
 * @brief Implements File Transfer Protocol
 *
 * @author Copyright (c) 1985, 1989 Regents of the University of California.All rights reserved.
 *
 * @note From: @(#)ftp.c  5.38 (Berkeley) 4/22/91
 *
 * */
char            copyright[] =
  "@(#) Copyright (c) 1985, 1989 Regents of the University of California.\n"
  "All rights reserved.\n";

#include <types.h>
#include "stdlib.h"

#include <sys/socket.h>

#include <posix_time_pub.h>     // nanosleep()

#include <unistd.h>
#include <errno.h>

#include <sys/stat.h>

#include <log_pub.h>

extern int      ftp_safeput;
extern int      ftp_verbose;
extern int      ftp_sendport;
extern int      ftp_debug;

#include "ftp_var.h"

extern int      control_fd;
extern int      data;

extern int      connected;      ///< connected to server 
extern int      code;           ///< return/reply code for ftp command 
extern int      crflag;         ///< if 1, strip car. rets. on ascii gets 
extern int      type;           ///< requested file transfer type 
extern int      curtype;        ///< current file transfer type 
extern char    *hostname;       ///< name of host connected to 
extern int      cpend;          ///< flag: if != 0, then pending server reply 
extern char     reply_string[];
extern char     lreply[];

char            typename[32];   ///< name of file transfer type 
int             unix_server;    ///< server is unix, can use binary for ascii 
int             ftp_port;       ///< htons'd port number for ftp service 

void            help(int argc, char *argv[]);

/**
 * @brief Init some variables: the ftp port...
 *
 * @param void
 *
 * @return Always 0
 *
 **/

static int do_init(void)
{
  ftp_port = 21;
  cpend = 0;                    ///< no pending replies 
  crflag = 1;                   ///< strip c.r. on ascii gets 
  ftp_sendport = -1;            ///< not using ports 

  //  (void) signal(SIGPIPE, lostpeer);
  return 0;
}

int ftp_close(void)
{

  struct timespec stTime;

  stTime.tv_sec = 1;
  stTime.tv_nsec = 0;

  if (!connected)
    return 1;
  command("QUIT");
  if (control_fd)
  {
    nanosleep(&stTime, NULL);
    shutdown(control_fd, SHUT_RDWR);
    close(control_fd);
  }
  control_fd = -1;
  connected = 0;
  data = -1;
  return 0;
}

int ftp_open(char *server, char *login, char *passwd)
{
  char           *host;
  unsigned short  port;

  LogPrint(ftpLogGet(), LOG_LEVEL_INFO, "Ftp::ftp_open %s %s %s\n", server, login, passwd);
  do_init();
  if (connected)
  {
    LogPrint(ftpLogGet(), LOG_LEVEL_INFO,
             "Ftp::ftp_open Already connected to %s, use close first\n", hostname);
    return -2;
  }
  port = ftp_port;
  host = hookup(server, port);

  if (host == NULL)
    return -3;

  connected = 1;
  /*
   * Set up defaults for FTP.
   */
  strcpy(typename, "binary");
  type = TYPE_I;
  curtype = TYPE_A;
  if (dologin(login, passwd))
  {
    ftp_close();
    return -4;
  }

#if defined(__unix__) && CHAR_BIT == 8

  /*
   * this ifdef is to keep someone form "porting" this to an incompatible
   * system and not checking this out. This way they have to think about it.
   */
  overbose = ftp_verbose;
  if (ftp_debug == 0)
    ftp_verbose = -1;
  if (command("SYST") == FTP_COMPLETE && overbose)
  {
    register char  *cp, c = 0;

    cp = index(reply_string + 4, ' ');
    if (cp == NULL)
      cp = index(reply_string + 4, '\r');
    if (cp)
    {
      if (cp[-1] == '.')
        cp--;
      c = *cp;
      *cp = '\0';
    }

    LogPrint(ftpLogGet(), LOG_LEVEL_INFO, "Ftp::ftp_open Remote system type is %s\n",
             reply_string + 4);
    if (cp)
      *cp = c;
  }
  if (!strncmp(reply_string, "215 UNIX Type: L8", 17))
  {
    unix_server = 1;
    /*
     * Set type to 0 (not specified by user),
     * meaning binary by default, but don't bother
     * telling server.  We can use binary
     * for text files unless changed by the user.
     */
    type = 0;
    strcpy(typename, "binary");
    if (overbose)
      LogPrint(ftpLogGet(), LOG_LEVEL_INFO, "Using %s mode to transfer files\n", typename);
  }
  else
  {
    unix_server = 0;
    if (overbose && !strncmp(reply_string, "215 TOPS20", 10))
      LogPrint(ftpLogGet(), LOG_LEVEL_INFO,
               "Remember to set tenex mode when transfering binary files from this machine.\n");
  }
  ftp_verbose = overbose;
#else
  //Unix auto-mode code skipped
#endif /* unix */
  return 0;
}

#if 0
int ftp_cd(char *directory)
{
  int             r;

  if (!strcmp("..", directory))
  {

    r = command("CDUP");
    if (r == FTP_ERROR && code == 500)
    {
      if (ftp_verbose)
        printf("CDUP command not recognized, trying XCUP\n");
      return command("XCUP") != FTP_COMPLETE;
    }
    return r != FTP_COMPLETE;
  }

  r = command("CWD %s", directory);
  if (r == FTP_ERROR && code == 500)
  {
    if (ftp_verbose)
      printf("CWD command not recognized, trying XCWD\n");
    return (command("XCWD %s", directory) != FTP_COMPLETE);
  }
  return r != FTP_COMPLETE;
}

int ftp_delete(char *file)
{
  return (command("DELE %s", file) != FTP_COMPLETE);
}

int ftp_mkdir(char *directory)
{
  int             r;

  r = command("MKD %s", directory);
  if (r == FTP_ERROR && code == 500)
  {
    if (ftp_verbose)
      printf("MKD command not recognized, trying XMKD\n");
    return (command("XMKD %s", directory) != FTP_COMPLETE);
  }
  return r != FTP_COMPLETE;
}

int ftp_rmdir(char *directory)
{
  int             r;

  r = command("RMD %s", directory);
  if (r == FTP_ERROR && code == 500)
  {
    if (ftp_verbose)
      printf("RMD command not recognized, trying XRMD\n");
    return (command("XRMD %s", directory) != FTP_COMPLETE);
  }
  return r != FTP_COMPLETE;
}
#endif

int ftp_rename(char *oldname, char *newname)
{
  int             r;
  char            cmd[MAX_BUF];

  strncpy(cmd, "RNFR ", MAX_BUF);
  strncpy(cmd + 5, oldname, MAX_BUF - 5);
  //  r = command("RNFR %s", oldname);
  r = command(cmd);
  if (r != FTP_CONTINUE)
    return 1;

  strncpy(cmd, "RNTO ", MAX_BUF);
  strncpy(cmd + 5, newname, MAX_BUF - 5);

  return (command(cmd) != FTP_COMPLETE);
}

int ftp_ls(int ll, char *pathname, int (*fct) (char *, unsigned))
{
  if (!connected)
  {
    LogPrint(ftpLogGet(), LOG_LEVEL_WARNING,
             "Ftp::ftp_ls Not conected to any server. Use open first\n");
    return -2;
  }

  return recvrequest(ll ? "LIST" : "NLST", pathname, fct, 80, 0);
}

/* low level file transfer functions */

int ftp_get(char *pathname, int (*fct) (char *, unsigned), unsigned bufsize, off_t rst)
{
  if (!connected)
  {
    LogPrint(ftpLogGet(), LOG_LEVEL_WARNING,
             "Ftp::ftp_get Not conected to any server. Use open first\n");
    return -2;
  }
  LogPrint(ftpLogGet(), LOG_LEVEL_INFO, "Ftp::ftp_get calling recvrequest with: %s %d %d\n",
           pathname, bufsize, rst);

  return recvrequest("RETR ", pathname, fct, bufsize, rst);

}

int ftp_put(char *pathname, int (*fct) (char *, unsigned), unsigned size, off_t rst, int append)
{
  char           *tmpname = NULL;
  int             r;
  char           *cmd = (append ? "APPE " : "STOR ");

  if (ftp_safeput)
  {
    tmpname = malloc(strlen(pathname) + 5);
    if (!tmpname)
      return 1;
    strcpy(tmpname, pathname);
    strcat(tmpname, ".tmp");
  }
  else
    tmpname = pathname;

  r = sendrequest(cmd, tmpname, fct, size, rst);
  if (r || !ftp_safeput)
    return r;
  r = ftp_rename(tmpname, pathname);
  if (tmpname != pathname)
    free(tmpname);
  return r;
}

#if 0
int ftp_putunique(char *pathname, char *uname, unsigned usize, int (*fct) (char *, unsigned),
                  unsigned bufsize)
{
  char           *b, *e;

  int             r = sendrequest("STOU", pathname, fct, bufsize, 0);

  if (uname != NULL && usize > 0)
  {
    /* getting the unique name */
    /* this works for wu-ftpd ... */
    b = lreply;
    while (*b != ':' && *b != 0)
      b++;
    if (b == 0)
      *uname = 0;
    else
    {
      e = b++;
      while (*e != ')' && *e != 0)
        e++;
      *e = 0;
      strncpy(uname, b, usize);
    }
  }
  return r;
}
#endif /* 0 */

/* high level file transfer functions */

FILE           *fout, *fin;

/**
 * @brief Writes a buffer
 *
 * @param buf The buffer
 * @param size The buffer size
 *
 * @return 0 on success,
 *         1 on failure.
 *
 * */

static int write_get_buf(char *buf, unsigned size)
{
  unsigned        nw = 0;

  if ((nw = write(fileno(fout), buf, size)) != size)
  {
    LogPrint(ftpLogGet(), LOG_LEVEL_ERROR, "Ftp::write_get_buf Error writing to local file\n");
    LogPrint(ftpLogGet(), LOG_LEVEL_INFO, "Ftp::write_get_buf errno = %d\n", errno);
    return 1;
  }
  return 0;
}

int ftp_getfile(char *remote_file, char *local_file, off_t rst)
{
  int             r;
  unsigned        bufsize;
  char            mode[3] = "r+";
  struct stat     st;

  if (!rst)
    strcpy(mode, "w");

  LogPrint(ftpLogGet(), LOG_LEVEL_INFO, "Ftp:ftp_getfile. Local file:%s Remote file:%s. Rst:%d\n",
           local_file, remote_file, rst);
  if ((fout = fopen(local_file, mode)) == NULL)
  {
    LogPrint(ftpLogGet(), LOG_LEVEL_ERROR, "Ftp::ftp_getfile  Cannot open local %s\n", local_file);
    return 2;
  }

  if (rst)
    if (fseek(fout, (long)rst, SEEK_SET) < 0)
      return 3;

  if (fstat(fileno(fout), &st) < 0 || st.st_blksize == 0)
    bufsize = BUFSIZ;
  else
    bufsize = st.st_blksize;
  bufsize = BUFSIZ;

  iprintf("ftp: bufsize:%d\n", bufsize);
  // All file transfers are binaries transfers
  type = TYPE_I;

  LogPrint(ftpLogGet(), LOG_LEVEL_INFO, "Ftp:ftp_getfile. Bufsize:%d\n", bufsize);
  r = ftp_get(remote_file, write_get_buf, bufsize, rst);
  fclose(fout);
  return r;
}

/**
 * @brief Reads a buffer
 *
 * @param buf The buffer
 * @param size The buffer size
 *
 * @return The bytes read
 *
 *
 * */

static int read_put_buf(char *buf, unsigned size)
{
  int             c = read(fileno(fin), buf, size);

  return c;
}

int ftp_putfile(char *local_file, char *remote_file, off_t rst, int append)
{
  int             r;
  unsigned        bufsize;
  struct stat     st;

  if ((fin = fopen(local_file, "r")) == NULL)
  {
    LogPrint(ftpLogGet(), LOG_LEVEL_ERROR, "Ftp::ftp_putfile Cannot open local file %s\n",
             local_file);
    return 2;
  }
  if (rst)
    if (fseek(fin, (long)rst, SEEK_SET) < 0)
      return 3;

  if (fstat(fileno(fin), &st) < 0 || st.st_blksize == 0)
    bufsize = BUFSIZ;
  else
    bufsize = st.st_blksize;
  iprintf("ftp: bufsize:%d\n", bufsize);
  //bufsize = BUFSIZ;
  // All file transfers are binaries transfers
  type = TYPE_I;

  r = ftp_put(remote_file, read_put_buf, bufsize, rst, append);
  fclose(fin);
  return r;
}

#if 0
int ftp_putfileunique(char *local_file, char *remote_file, char *uname, unsigned usize)
{
  int             r;
  unsigned        bufsize;
  struct stat     st;

  if ((fin = fopen(local_file, "r")) == NULL)
  {
    fprintf(stderr, "Cannot open local file %s\n", local_file);
    return 2;
  }

  if (fstat(fileno(fin), &st) < 0 || st.st_blksize == 0)
    bufsize = BUFSIZ;
  else
    bufsize = st.st_blksize;

  r = ftp_putunique(remote_file, uname, usize, read_put_buf, bufsize);
  fclose(fin);
  return r;
}

int ftp_sendcommand(char *cmd, char *buf, unsigned size)
{
  int             r = command(cmd);

  memcpy(buf, lreply, ((size) < (BUFSIZ) ? size : BUFSIZ));
  return r != FTP_COMPLETE;
}
#endif /* 0 */

// ---

/*******************************************************************************/

/**
 * @}
 * */
