/* setjmp/longjmp functions for Xtensa.  */

/*
 * Copyright (c) 2001, 2002 by Tensilica Inc.  ALL RIGHTS RESERVED.
 * These coded instructions, statements, and computer programs are the
 * copyrighted works and confidential proprietary information of Tensilica Inc.
 * They may not be modified, copied, reproduced, distributed, or disclosed to
 * third parties in any manner, medium, or form, in whole or in part, without
 * the prior written consent of Tensilica Inc.
 */

/*
   Windowed ABI:	
	
   This implementation relies heavily on the Xtensa register window
   mechanism.  Setjmp flushes all the windows except its own to the
   stack and then copies registers from the save areas on the stack
   into the jmp_buf structure, along with the special register state
   and the return address of the call to setjmp.  Longjmp invalidates
   all the windows except its own, restores the special registers, and
   then sets things up so that it will return to the right place,
   using a window underflow to automatically restore the registers.

   Note that it would probably be sufficient to only copy the
   registers from setjmp's caller into jmp_buf.  However, we also copy
   the save area located at the stack pointer of setjmp's caller.
   This save area will typically remain intact until the longjmp call.
   The one exception is when there is an intervening alloca in
   setjmp's caller.  This is certainly an unusual situation and is
   likely to cause problems in any case (the storage allocated on the
   stack cannot be safely accessed following the longjmp).  As bad as
   it is, on most systems this situation would not necessarily lead to
   a catastrophic failure.  If we did not preserve the extra save area
   on Xtensa, however, it would.  When setjmp's caller returns after a
   longjmp, there will be a window underflow; an invalid return
   address or stack pointer in the save area will almost certainly
   lead to a crash.  Keeping a copy of the extra save area in the
   jmp_buf avoids this with only a small additional cost.  If setjmp
   and longjmp are ever time-critical, this could be removed.  
	

   Call0 ABI:	
	
   Much like other abis, just saves the necessary registers to the stack 
   and restores them later. Much less needs to be done
*/
   
#include <xtensa/coreasm.h>
/* change by Hareesh for T1050.5 XCC port */
#if 1
#define MIN_ESA  0
#define M0 MR_0
#define M1 MR_1
#define M2 MR_2
#define M3 MR_3
#endif
/* end of change by Hareesh */

#define SYS_nop	0


#if XCHAL_HAVE_WINDOWED && !__XTENSA_CALL0_ABI__

/*
 *  int  setjmp( jmp_buf env );
 */
	.text
	.align	4
	.global	setjmp
	.type	setjmp,@function
setjmp:
	entry	sp, 16+MIN_ESA

	/* flush registers */
	mov	a4, a2			# save a2 (jmp_buf)
	movi	a2, SYS_nop
	syscall
	mov	a2, a4			# restore a2

	/* copy the register save area at (sp - 16) */
	addi	a5, a1, -16
	l32i	a3, a5, 0
	l32i	a4, a5, 4
	s32i	a3, a2, 0
	s32i	a4, a2, 4
	l32i	a3, a5, 8
	l32i	a4, a5, 12 
	s32i	a3, a2, 8
	s32i	a4, a2, 12
	
	/* copy 0-8 words from the register overflow area */
	extui	a3, a0, 30, 2
	blti	a3, 2, .Lendsj
	l32i	a7, a1, 4+MIN_ESA
	slli	a4, a3, 4
	sub	a5, a7, a4
	addi	a6, a2, 16
	addi	a7, a7, -16		# a7 = end of register overflow area
.Lsjloop:
	l32i	a3, a5, 0
	l32i	a4, a5, 4
	s32i	a3, a6, 0
	s32i	a4, a6, 4
	l32i	a3, a5, 8
	l32i	a4, a5, 12
	s32i	a3, a6, 8
	s32i	a4, a6, 12
	addi	a5, a5, 16
	addi	a6, a6, 16
	blt	a5, a7, .Lsjloop
.Lendsj:

	/* copy the register save area at (sp) = (sp + 16 - 16) */
	l32i	a3, a1, 0+MIN_ESA
	l32i	a4, a1, 4+MIN_ESA
	s32i	a3, a2, 48
	s32i	a4, a2, 52
	l32i	a3, a1, 8+MIN_ESA
	l32i	a4, a1, 12+MIN_ESA 
	s32i	a3, a2, 56
	s32i	a4, a2, 60

	/* save the return address, including the window size bits */
	s32i	a0, a2, 64
	
	/* save the special registers */
#if XCHAL_HAVE_LOOPS
	rsr	a3, SAR
	rsr	a4, LCOUNT
	s32i	a3, a2, 68
	s32i	a4, a2, 72

	rsr	a3, LBEG
	rsr	a4, LEND
	s32i	a3, a2, 76
	s32i	a4, a2, 80

#else /* !XCHAL_HAVE_LOOPS */
	rsr	a3, SAR
	movi	a4, 0
	s32i	a3, a2, 68
	s32i	a4, a2, 72

	s32i	a4, a2, 76
	s32i	a4, a2, 80
#endif /* !XCHAL_HAVE_LOOPS */

#if XCHAL_HAVE_MAC16
	rsr	a3, ACCLO
	rsr	a4, ACCHI
	s32i	a3, a2, 84
	s32i	a4, a2, 88

	rsr	a3, M0
	rsr	a4, M1
	s32i	a3, a2, 92
	s32i	a4, a2, 96

	rsr	a3, M2
	rsr	a4, M3
	s32i	a3, a2, 100
	s32i	a4, a2, 104
#endif /* XCHAL_HAVE_MAC16 */
	movi	a2, 0
	retw
	.size	setjmp,.-setjmp



/*
 *  void  longjmp( jmp_buf env, int val );
 */
	.align	4
	.global	longjmp
	.type	longjmp,@function
longjmp:
	entry	sp, 16+MIN_ESA
	/*  a2 == &env, a3 == val  */

#if XCHAL_MAYHAVE_ERRATUM_XEA1KWIN
# define AR_WB	a15	/* using this register triggers early any overflow that a kernel-mode level-one interrupt might otherwise cause */
#else
# define AR_WB	a5	/* using this register is more efficient, triggers less overflows */
#endif
	/* Invalidate all but the current window;
	   set WindowStart to (1 << WindowBase). */
	rsr	AR_WB, WINDOWBASE
	movi	a4, 1
	ssl	AR_WB
	sll	a4, a4
	wsr	a4, WINDOWSTART
	rsync
	
	/* Return to the return address of the setjmp, using the
	   window size bits from the setjmp call so that the caller
	   will be able to find the return value that we put in a2. */

	l32i	a0, a2, 64

	/* Copy the first 4 saved registers from jmp_buf into the save area
	   at the current sp so that the values will be restored to registers
	   when longjmp returns. */

	addi	a7, a1, -16
	l32i	a4, a2, 0
	l32i	a5, a2, 4
	s32i	a4, a7, 0
	s32i	a5, a7, 4
	l32i	a4, a2, 8
	l32i	a5, a2, 12
	s32i	a4, a7, 8
	s32i	a5, a7, 12

	/* Copy the remaining 0-8 saved registers. */
	extui	a7, a0, 30, 2
	blti	a7, 2, .Lendlj
	l32i	a8, a2, 52
	slli	a4, a7, 4
	sub	a6, a8, a4
	addi	a5, a2, 16
	addi	a8, a8, -16		# a8 = end of register overflow area
.Lljloop:
	l32i	a7, a5, 0
	l32i	a4, a5, 4
	s32i	a7, a6, 0
	s32i	a4, a6, 4
	l32i	a7, a5, 8
	l32i	a4, a5, 12
	s32i	a7, a6, 8
	s32i	a4, a6, 12
	addi	a5, a5, 16
	addi	a6, a6, 16
	blt	a6, a8, .Lljloop
.Lendlj:
	
	/* The 4 words saved from the register save area at the target's
	   sp are copied back to the target procedure's save area.  The
	   only point of this is to prevent a catastrophic failure in
	   case the contents were moved by an alloca after calling
	   setjmp.  This is a bit paranoid but it doesn't cost much. */

	l32i	a7, a2, 4		# load the target stack pointer
	addi	a7, a7, -16		# find the destination save area
	l32i	a4, a2, 48
	l32i	a5, a2, 52
	s32i	a4, a7, 0
	s32i	a5, a7, 4
	l32i	a4, a2, 56
	l32i	a5, a2, 60
	s32i	a4, a7, 8
	s32i	a5, a7, 12

	/* Restore the special registers. */
#if XCHAL_HAVE_LOOPS
	l32i	a4, a2, 68
	l32i	a5, a2, 72
	wsr	a4, SAR
	wsr	a5, LCOUNT

	l32i	a4, a2, 76
	l32i	a5, a2, 80
	wsr	a4, LBEG
	wsr	a5, LEND

#else /* !XCHAL_HAVE_LOOPS */
	l32i	a4, a2, 68
	wsr	a4, SAR
#endif /* !XCHAL_HAVE_LOOPS */

#if XCHAL_HAVE_MAC16
	l32i	a4, a2, 84
	l32i	a5, a2, 88
	wsr	a4, ACCLO
	wsr	a5, ACCHI

	l32i	a4, a2, 92
	l32i	a5, a2, 96
	wsr	a4, M0
	wsr	a5, M1

	l32i	a4, a2, 100
	l32i	a5, a2, 104
	wsr	a4, M2
	wsr	a5, M3

#endif /* XCHAL_HAVE_MAC16 */
	/* wait for the special registers to be updated */
	isync
	/*
	 * NOTE: according to GNATS 6362 (erratum "XEA1KWIN"), XEA1 code
	 * must have at least one instruction between ISYNC and RETW if
	 * it is ever called in kernel vector mode with interrupts enabled.
	 */

	/* return v ? v : 1 */
	movi	a2, 1
	movnez	a2, a3, a3

	retw
	.size	longjmp,.-longjmp

#else
	.text
	.align	4
	.global	setjmp
	.type	setjmp,@function
setjmp:	
	s32i	a11, a2, 0
	s32i	a12, a2, 4
	s32i	a13, a2, 8
	s32i	a14, a2, 12
	s32i	a15, a2, 16
	movi	a2, 0
	ret

	.align	4
	.global	longjmp
	.type	longjmp,@function
longjmp:
	l32i	a11, a2, 0
	l32i	a12, a2, 4
	l32i	a13, a2, 8
	l32i	a14, a2, 12
	l32i	a15, a2, 16

	/* return v ? v : 1 */
	movi	a2, 1
	movnez	a2, a3, a3

	ret
	.size	longjmp,.-longjmp

#endif
