/* ANSI C standard library function strcmp.  */

/*
 * Copyright (c) 2001,2003 by Tensilica Inc.  ALL RIGHTS RESERVED.
 * These coded instructions, statements, and computer programs are the
 * copyrighted works and confidential proprietary information of Tensilica Inc.
 * They may not be modified, copied, reproduced, distributed, or disclosed to
 * third parties in any manner, medium, or form, in whole or in part, without
 * the prior written consent of Tensilica Inc.
 */

#include <xtensa/coreasm.h>
/* change by Hareesh for T1050.5 XCC port */
#if 1
#define MIN_ESA  0
#define XCHAL_HAVE_L32R                 1       /* 1 if L32R is available for loading constants */
#endif
/* end of change by Hareesh */

#if XCHAL_HAVE_BE
#define	MASK0 0xff000000
#define	MASK1 0x00ff0000
#define	MASK2 0x0000ff00
#define	MASK3 0x000000ff
#else
#define	MASK0 0x000000ff
#define	MASK1 0x0000ff00
#define	MASK2 0x00ff0000
#define	MASK3 0xff000000
#endif

#define MASK4 0x40404040

#if XCHAL_HAVE_L32R
	.literal .Lmask0, MASK0
	.literal .Lmask1, MASK1
	.literal .Lmask2, MASK2
	.literal .Lmask3, MASK3
	.literal .Lmask4, MASK4
#endif /* XCHAL_HAVE_L32R */

/*
 * Register use
 *   a0/ return address
 *   a1/ stack pointer
 *   a2/ s1
 *   a3/ s2
 *   a4/ mask0 
 *   a5/ mask1
 *   a6/ mask2
 *   a7/ mask3 or mask4
 *   a8/ tmp
 *   a9/ tmp
 *   a10/ tmp
 */

	.text
	.align	4
	.global	strcmp
	.type	strcmp,@function
strcmp:
#if !XCHAL_HAVE_WINDOWED || __XTENSA_CALL0_ABI__
	addi a1, a1, -16+MIN_ESA
	s32i a10, a1, 8
#else
	entry a1, 16+MIN_ESA
#endif
	/* a2/ s1, a3/ s2 */

	l8ui	a8, a2, 0	# byte 0 from s1
	l8ui	a9, a3, 0	# byte 0 from s2
	movi	a10, 3		# mask
	bne 	a8, a9, .Lretdiff

        or	a11, a2, a3
        bnone   a11, a10, .Laligned

	xor	a11, a2, a3	# compare low two bits of s1 and s2
	bany	a11, a10, .Lunaligned	# if s1 and s2 have different alignment

	/* s1/s2 are not word-aligned */
	addi	a2, a2, 1	# advance s1
	beqz	a8, .Leq	# bytes equal, if zero, strings are equal
	addi	a3, a3, 1	# advance s2
	bnone	a2, a10, .Laligned # if s1/s2 now aligned
	l8ui	a8, a2, 0	# byte 1 from s1
	l8ui	a9, a3, 0	# byte 1 from s2
	addi	a2, a2, 1	# advance s1
	bne	a8, a9, .Lretdiff # if different, return difference
	beqz	a8, .Leq	# bytes equal, if zero, strings are equal
	addi	a3, a3, 1	# advance s2
	bnone	a2, a10, .Laligned # if s1/s2 now aligned
	l8ui	a8, a2, 0	# byte 2 from s1
	l8ui	a9, a3, 0	# byte 2 from s2
	addi	a2, a2, 1	# advance s1
	bne	a8, a9, .Lretdiff # if different, return difference
	beqz	a8, .Leq	# bytes equal, if zero, strings are equal
	addi	a3, a3, 1	# advance s2
	j	.Laligned

/*
 * s1 and s2 have different alignment
 *
 * If the zero-overhead loop option is available, use an (almost)
 * infinite zero-overhead loop with conditional exits so we only pay
 * for taken branches when exiting the loop
 *
 * Note: It is important for this unaligned case to come before the
 * code for aligned strings, because otherwise some of the branches
 * above cannot reach and have to be transformed to branches around
 * jumps.  The unaligned code is smaller and the branches can reach
 * over it.
 */

	.align	4
#if XCHAL_HAVE_LOOPS
#if XCHAL_HAVE_DENSITY
	# (2 mod 4) alignment for loop instruction
#else
	# (1 mod 4) alignment for loop instruction
	.byte	0
	.byte	0
#endif /* XCHAL_HAVE_DENSITY */
#endif /* XCHAL_HAVE_LOOPS */
.Lunaligned:
#if XCHAL_HAVE_LOOPS
#if XCHAL_HAVE_DENSITY
	_movi.n	a8, 0		# set up for the maximum loop count
#else
	_movi	a8, 0		# set up for the maximum loop count
#endif /* XCHAL_HAVE_DENSITY */
	loop	a8, .Lretdiff	# loop forever (almost anyway)
#endif /* XCHAL_HAVE_LOOPS */
.Lnextbyte:
	l8ui	a8, a2, 0
	l8ui	a9, a3, 0
	addi	a2, a2, 1
	bne	a8, a9, .Lretdiff
	addi	a3, a3, 1
#if XCHAL_HAVE_LOOPS
	beqz	a8, .Lretdiff
#else /* !XCHAL_HAVE_LOOPS */
	bnez	a8, .Lnextbyte
#endif /* !XCHAL_HAVE_LOOPS */
.Lretdiff:
	sub	a2, a8, a9
#if !XCHAL_HAVE_WINDOWED
	l32i a10, a1, 8
	addi a1, a1, 16
	ret
#else
	retw
#endif

/*
 * s1 is word-aligned, s2 is word-aligned
 *
 * If the zero-overhead loop option is available, use an (almost)
 * infinite zero-overhead loop with conditional exits so we only pay
 * for taken branches when exiting the loop
 */

/* New algorithm: rely on the fact that all normal ascii is between 32 and 127
 * Rather than check all bytes for zero, 
 * Take one word (4 bytes) call it w1
 * Shift w1 left by one into w1' 
 * Or w1 and w1', for all normal ascii bit 6 will be 1, for zero it won't
 * Check that all 4 bit 6's (one for each byte) are one, 
 *  if they are, we are definitely not done
 *  if they are not, we are probably done, but need to check for zero
 */
	.align	4
#if XCHAL_HAVE_LOOPS
#if !XCHAL_HAVE_L32R
	# (2 mod 4) alignment for loop instruction
	.byte	0
	.byte	0
#endif /* !XCHAL_HAVE_L32R */
.Laligned:
/* change by Hareesh for T1050.5 XCC port */
#if 0 /* RA.2 */
	.begin no-transform
#else /* T1050.5 */
	.begin no-generics
#endif
/* end of change */
#if XCHAL_HAVE_L32R
	l32r	a4, .Lmask0	# mask for byte 0
	l32r	a7, .Lmask4	
#else
	const16	a4, MASK0@h
	const16	a4, MASK0@l
	const16	a7, MASK4@h
	const16	a7, MASK4@l
#endif
        loop    a4, .Laligned_done	# loop forever (a4 is more than than
					# the maximum number of iterations) 

	l32i	a8, a2, 0	# get word from s1
	l32i	a9, a3, 0	# get word from s2
        slli	a5, a8, 1
	bne	a8, a9, .Lwne2
        or	a9, a8, a5
        bnall   a9, a7, .Lprobeq

	l32i	a8, a2, 4	# get word from s1
	l32i	a9, a3, 4	# get word from s2
        slli	a5, a8, 1
	bne	a8, a9, .Lwne2
        or	a9, a8, a5
        bnall   a9, a7, .Lprobeq2
	addi	a2, a2, 8	# advance s1 pointer
	addi	a3, a3, 8	# advance s2 pointer
.Laligned_done:
        or	a1, a1, a1	# nop

#else /* !XCHAL_HAVE_LOOPS */

.Laligned:
	movi	a4, MASK0	# mask for byte 0
	movi	a7, MASK4
	j	.Lfirstword
.Lnextword:
	addi	a2, a2, 4	# advance s1 pointer
	addi	a3, a3, 4	# advance s2 pointer
.Lfirstword:
	l32i	a8, a2, 0	# get word from s1
	l32i	a9, a3, 0	# get word from s2
	slli	a5, a8, 1
	bne	a8, a9, .Lwne2
	or	a9, a8, a5
	ball	a9, a7, .Lnextword
#endif /* !XCHAL_HAVE_LOOPS */

	# align	(0 mod 4)
#if XCHAL_HAVE_LOOPS
.Lprobeq2:
	addi	a2, a2, 4	# advance s1 pointer
	addi	a3, a3, 4	# advance s2 pointer
#endif
.Lprobeq:  
          /* words are probably equal, but check for sure */
          /*  if not, loop over the rest of string using normal algorithm */

	bnone	a8, a4, .Leq	# if byte 0 is zero
#if XCHAL_HAVE_L32R
	l32r	a5, .Lmask1	# mask for byte 1
	l32r	a6, .Lmask2	# mask for byte 2
	bnone	a8, a5, .Leq	# if byte 1 is zero
	l32r	a7, .Lmask3	# mask for byte 3
	bnone	a8, a6, .Leq	# if byte 2 is zero
	bnone	a8, a7, .Leq	# if byte 3 is zero
	# align (1 mod 4)
#else
	const16	a5, MASK1@h	# mask for byte 1
	const16	a5, MASK1@l
	bnone	a8, a5, .Leq	# if byte 1 is zero
	const16	a6, MASK2@h	# mask for byte 2
	const16	a6, MASK2@l
	bnone	a8, a6, .Leq	# if byte 2 is zero
	const16	a7, MASK3@h	# mask for byte 3
	const16	a7, MASK3@l
	bnone	a8, a7, .Leq	# if byte 3 is zero
	# align (2 mod 4)
#endif /* XCHAL_HAVE_L32R */
#if XCHAL_HAVE_DENSITY
	addi.n	a2, a2, 4	# advance s1 pointer
	addi.n	a3, a3, 4	# advance s2 pointer
	# align (1 mod 4) or (2 mod 4)
#else
	addi	a2, a2, 4	# advance s1 pointer
	addi	a3, a3, 4	# advance s2 pointer
	or	a1, a1, a1	# nop
#if !XCHAL_HAVE_L32R
	or	a1, a1, a1	# nop
#endif
	# align (2 mod 4)
#endif /* XCHAL_HAVE_DENSITY */
#if XCHAL_HAVE_LOOPS

        loop    a4, .Leq    	# loop forever (a4 is bigger than max iters) 
/* change by Hareesh for T1050.5 XCC port */
#if 0 /* RA.2 */
	.end no-transform
#else /* T1050.5 */
	.end no-generics
#endif
/* end of change */

	l32i	a8, a2, 0	# get word from s1
	l32i	a9, a3, 0	# get word from s2
	addi	a2, a2, 4	# advance s1 pointer
	bne	a8, a9, .Lwne
	bnone	a8, a4, .Leq	# if byte 0 is zero
	bnone	a8, a5, .Leq	# if byte 1 is zero
	bnone	a8, a6, .Leq	# if byte 2 is zero
	bnone	a8, a7, .Leq	# if byte 3 is zero
	addi	a3, a3, 4	# advance s2 pointer

#else /* !XCHAL_HAVE_LOOPS */

	j	.Lfirstword2
.Lnextword2:
	addi	a3, a3, 4	# advance s2 pointer
.Lfirstword2:
	l32i	a8, a2, 0	# get word from s1
	l32i	a9, a3, 0	# get word from s2
	addi	a2, a2, 4	# advance s1 pointer
	bne	a8, a9, .Lwne
	bnone	a8, a4, .Leq	# if byte 0 is zero
	bnone	a8, a5, .Leq	# if byte 1 is zero
	bnone	a8, a6, .Leq	# if byte 2 is zero
	bany	a8, a7, .Lnextword2	# if byte 3 is zero
#endif /* !XCHAL_HAVE_LOOPS */

.Leq:	/* words are equal, some byte is zero */
	movi	a2, 0		# return equal
#if !XCHAL_HAVE_WINDOWED
	l32i a10, a1, 8
	addi a1, a1, 16
	ret
#else
	retw
#endif

.Lwne2:	/* Words are not equal.  On big-endian processors, if none of the
	   bytes are zero, the return value can be determined by a simple
	   comparison.  */
#ifdef __XTENSA_EB__
        or	a10, a8, a5
        bnall   a10, a7, .Lsomezero
	bgeu	a8, a9, .Lposreturn 
	movi	a2, -1
	retw
.Lposreturn:
	movi	a2, 1
	retw
.Lsomezero:	#there is probably some zero byte
#else
#endif
.Lwne:	/* words are not equal */
	xor	a2, a8, a9	# get word with nonzero in byte that differs
	bany	a2, a4, .Ldiff0	# if byte 0 differs
	movi	a5, MASK1	# mask for byte 1
	bnone	a8, a4, .Leq	# if byte 0 is zero
	bany	a2, a5, .Ldiff1	# if byte 1 differs
	movi	a6, MASK2	# mask for byte 2
	bnone	a8, a5, .Leq	# if byte 1 is zero
	bany	a2, a6, .Ldiff2	# if byte 2 differs
	bnone	a8, a6, .Leq	# if byte 2 is zero
#ifdef __XTENSA_EB__
.Ldiff3:
.Ldiff2:
.Ldiff1:
	/*
	 * byte 0 is equal (at least)
	 * and there is a difference before a zero byte
	 * just subtract words to get return value
	 * the high order equal bytes cancel, leaving room for the sign
	 */
	sub	a2, a8, a9
#if !XCHAL_HAVE_WINDOWED
	l32i a10, a1, 8
	addi a1, a1, 16
	ret
#else
	retw
#endif
.Ldiff0:
	/* need to make room for the sign, so can't subtract whole words */
	extui	a10, a8, 24, 8
	extui	a11, a9, 24, 8
	sub	a2, a10, a11
#if !XCHAL_HAVE_WINDOWED
	l32i a10, a1, 8
	addi a1, a1, 16
	ret
#else
	retw
#endif
#else /* !__XTENSA_EB__ */
	/* little-endian is a little more difficult because can't subtract
	   whole words */
.Ldiff3:
	/* bytes 0-2 are equal, byte 3 is different */
	/* for little-endian need to have a sign bit for the difference */
	extui	a10, a8, 24, 8
	extui	a11, a9, 24, 8
	sub	a2, a10, a11
#if !XCHAL_HAVE_WINDOWED
	l32i a10, a1, 8
	addi a1, a1, 16
	ret
#else
	retw
#endif
.Ldiff0:
	/* byte 0 is different */
	extui	a10, a8, 0, 8
	extui	a11, a9, 0, 8
	sub	a2, a10, a11
#if !XCHAL_HAVE_WINDOWED
	l32i a10, a1, 8
	addi a1, a1, 16
	ret
#else
	retw
#endif
.Ldiff1:
	/* byte 0 is equal, byte 1 is different */
	extui	a10, a8, 8, 8
	extui	a11, a9, 8, 8
	sub	a2, a10, a11
#if !XCHAL_HAVE_WINDOWED
	l32i a10, a1, 8
	addi a1, a1, 16
	ret
#else
	retw
#endif
.Ldiff2:
	/* bytes 0-1 are equal, byte 2 is different */
	extui	a10, a8, 16, 8
	extui	a11, a9, 16, 8
	sub	a2, a10, a11
#if !XCHAL_HAVE_WINDOWED
	l32i a10, a1, 8
	addi a1, a1, 16
	ret
#else
	retw
#endif
#endif /* !__XTENSA_EB */
