/* ANSI C standard library function strlen.  */

/*
 * Copyright (c) 2001,2003 by Tensilica Inc.  ALL RIGHTS RESERVED.
 * These coded instructions, statements, and computer programs are the
 * copyrighted works and confidential proprietary information of Tensilica Inc.
 * They may not be modified, copied, reproduced, distributed, or disclosed to
 * third parties in any manner, medium, or form, in whole or in part, without
 * the prior written consent of Tensilica Inc.
 */

#include <xtensa/coreasm.h>
/* change by Hareesh for T1050.5 XCC port */
#if 1
#define MIN_ESA 0
/*----------------------------------------------------------------------
 *  abi_entry
 *
 *  generates proper entry instruction for either abi.
 *  limitation: doesn't increment the stack pointer for call0 abi. If
 *  stack manipulation is necessary, then you need to do it by hand
 *
 */

          .macro abi_entry reg, size
#if XCHAL_HAVE_WINDOWED
          entry \reg, \size
#else
          /* do nothing */
#endif
          .endm

/*----------------------------------------------------------------------
 *  abi_return
 *
 *  generates proper return instruction for either abi.
 *  limitation: doesn't restore the stack pointer for the call0
 *  abi. You must do that by hand
 *
 */

          .macro abi_return
#if XCHAL_HAVE_WINDOWED
          retw
#else
          ret
#endif
          .endm
#endif
/* end of change by Hareesh */

#if XCHAL_HAVE_BE
#define	MASK0 0xff000000
#define	MASK1 0x00ff0000
#define	MASK2 0x0000ff00
#define	MASK3 0x000000ff
#else
#define	MASK0 0x000000ff
#define	MASK1 0x0000ff00
#define	MASK2 0x00ff0000
#define	MASK3 0xff000000
#endif

	.text
	.align	4
	.global	strlen
	.type	strlen,@function
strlen:
	abi_entry a1, 16+MIN_ESA
	# a2/ s
	addi	a3, a2, -4	# because we overincrement at the end
	movi	a4, MASK0	# mask for byte 0
	movi	a5, MASK1	# mask for byte 1
	movi	a6, MASK2	# mask for byte 2
	movi	a7, MASK3	# mask for byte 3
	bbsi.l	a2, 0, .L1mod2	# if only  8-bit aligned
	bbsi.l	a2, 1, .L2mod4	# if only 16-bit aligned
	j	.Laligned

/*
 * String is word-aligned.
 */
	.align	4
#if XCHAL_HAVE_LOOPS
#if XCHAL_HAVE_DENSITY
	# (2 mod 4) alignment for loop instruction
#else
	# (1 mod 4) alignment for loop instruction
	.byte	0
	.byte	0
#endif /* XCHAL_HAVE_DENSITY */
#endif /* XCHAL_HAVE_LOOPS */
.Laligned:
#if XCHAL_HAVE_LOOPS
#if XCHAL_HAVE_DENSITY
	_movi.n	a8, 0		# set up for the maximum loop count
#else
	_movi	a8, 0		# set up for the maximum loop count
#endif /* XCHAL_HAVE_DENSITY */
	loop	a8, .Lz3	# loop forever (almost anyway)
#endif /* XCHAL_HAVE_LOOPS */
	l32i	a8, a3, 4	# get next word of string
	addi	a3, a3, 4	# advance string pointer
	bnone	a8, a4, .Lz0	# if byte 0 is zero
	bnone	a8, a5, .Lz1	# if byte 1 is zero
	bnone	a8, a6, .Lz2	# if byte 2 is zero
#if XCHAL_HAVE_LOOPS
	bnone	a8, a7, .Lz3	# if byte 3 is zero
#else /* !XCHAL_HAVE_LOOPS */
	bany	a8, a7, .Laligned	# repeat if byte 3 is non-zero
#endif /* !XCHAL_HAVE_LOOPS */

.Lz3:	# byte 3 is zero
	addi	a3, a3, 3	# point to zero byte
.Lz0:	# byte 0 is zero
	sub	a2, a3, a2	# subtract to get length
	abi_return
.Lz1:	# byte 1 is zero
	addi	a3, a3, 1	# point to zero byte
	sub	a2, a3, a2	# subtract to get length
	abi_return
.Lz2:	# byte 2 is zero
	addi	a3, a3, 2	# point to zero byte
	sub	a2, a3, a2	# subtract to get length
	abi_return

.L1mod2:	# address is odd
	l8ui	a8, a3, 4	# get byte 0
	addi	a3, a3, 1	# advance string pointer
	beqz	a8, .Lz3	# if byte 0 is zero
	bbci.l	a3, 1, .Laligned # if string pointer is now word-aligned

.L2mod4:	# address is 2 mod 4
	addi	a3, a3, 2	# advance ptr for aligned access
	l32i	a8, a3, 0	# get word with first two bytes of string
	bnone	a8, a6, .Lz2	# if byte 2 (of word, not string) is zero
	bany	a8, a7, .Laligned # if byte 3 (of word, not string) is nonzero
	# byte 3 is zero
	addi	a3, a3, 3	# point to zero byte
	sub	a2, a3, a2	# subtract to get length
	abi_return
