/*
 * _atexit: This is a simplified version of the standard atexit function.
 * It is only intended to be used by crt0 to register the _fini function
 * for code in the ELF finalization section.  Using the standard version
 * requires that all programs link in malloc, leading to a significant
 * increase in code size for programs that would otherwise not need malloc.
 */

#include <stddef.h>
#include <stdlib.h>
#include <reent.h>

/*
 * Register a function to be performed at exit.
 */

int
_DEFUN (_atexit,
	(fn),
	_VOID _EXFUN ((*fn), (_VOID)))
{
  register struct _atexit *p;

#ifndef _REENT_SMALL
  if ((p = _REENT->_atexit) == NULL)
    _REENT->_atexit = p = &_REENT->_atexit0;
#else
  p = &_REENT->_atexit;
#endif
  if (p->_ind >= _ATEXIT_SIZE)
    return -1;
  p->_fns[p->_ind++] = fn;
  return 0;
}

