
/*
 * Copyright 2007 Design of Systems on Silicon (DS2) All rights reserved.
 */

/**
 * @addtogroup zlib
 * @{
**/

/**
 * @file zlib_pub.h
 * @brief GZIP decompression function
 * 
 * @internal
 * 
 * @author fblasco, vromero, GNU source
 * @date 18/09/2002
 * @todo Make this code re-entrant
 * 
**/

#ifndef _ZLIB_PUB_H_
#  define _ZLIB_PUB_H_

/*
 *******************************************************************************
 ** Global constants 
 *******************************************************************************
 */

/*
 *******************************************************************************
 ** Global macros 
 *******************************************************************************
 */

/*
 *******************************************************************************
 ** Global data types 
 *******************************************************************************
 */

/*
 *******************************************************************************
 ** Global variables (declaration) 
 *******************************************************************************
 */

/*
 *******************************************************************************
 ** Global functions (prototypes) 
 *******************************************************************************
 */

/**
 * @brief This function uncompress data using the "deflate" method of the program GZIP.
 *
 * @param window  points to start of output window (where file will be decompressed)
 * 
 * @param window_size  size of output window, must be at least as big as output (decompressed) 
 *                     file expected. WARNING!!! This size must be a power of two.
 *                     
 * @param malloc_memory points to free memory space that can be used by decompressor 
 *
 * @param malloc_memory_size size of free memory space. Recommended at least 16KBytes     
 *                           WARNING!!! This size must be a power of two.
 *                                    
 * @param get_byte points to function which returns input bytes (from compressed file)
 * 
 * @param unget_byte 
 * 
 * @return Number of output bits or -1U (0xFFFFFFFFU) if error in decompression
 **/
INT32U          ZUncompress(INT8U *window, INT32U window_size, INT8U *malloc_memory,
                            INT32U malloc_memory_size, INT8U (*get_byte) (void),
                            void (*unget_byte) (void));

#endif // _ZLIB_PUB_H_

/**
 * @}
**/
